import macros from '../engine-javascript/compile/macros.js';
import macros$1 from '../engine-assemblyscript/compile/macros.js';
import compileToJavascript from '../engine-javascript/compile/index.js';
import compileToAssemblyscript from '../engine-assemblyscript/compile/index.js';
import { generate, attachInletCallers, attachOutletListeners } from './code-variable-names.js';
import { buildGraphTraversalDeclare, buildGraphTraversalLoop, preCompileSignalAndMessageFlow } from './compile-helpers.js';
import { trimGraph } from '../dsp-graph/traversal.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
var index = (graph, nodeImplementations, settings) => {
    const { audioSettings, arrays, inletCallerSpecs, outletListenerSpecs, target, debug, } = validateSettings(settings);
    const macros = getMacros(target);
    const codeVariableNames = generate(nodeImplementations, graph, debug);
    attachInletCallers(codeVariableNames, inletCallerSpecs);
    attachOutletListeners(codeVariableNames, outletListenerSpecs);
    const graphTraversalDeclare = buildGraphTraversalDeclare(graph, inletCallerSpecs);
    const graphTraversalLoop = buildGraphTraversalLoop(graph);
    const trimmedGraph = trimGraph(graph, graphTraversalDeclare);
    return {
        status: 0,
        code: executeCompilation({
            target,
            graph: trimmedGraph,
            graphTraversalDeclare,
            graphTraversalLoop,
            nodeImplementations,
            audioSettings,
            arrays,
            inletCallerSpecs,
            outletListenerSpecs,
            codeVariableNames,
            macros,
            debug,
            precompiledPortlets: {
                precompiledInlets: {},
                precompiledOutlets: {},
            },
        }),
    };
};
/** Asserts settings are valid (or throws error) and sets default values. */
const validateSettings = (settings) => {
    const arrays = settings.arrays || {};
    const inletCallerSpecs = settings.inletCallerSpecs || {};
    const outletListenerSpecs = settings.outletListenerSpecs || {};
    const debug = settings.debug || false;
    if (![32, 64].includes(settings.audioSettings.bitDepth)) {
        throw new InvalidSettingsError(`"bitDepth" can be only 32 or 64`);
    }
    return {
        ...settings,
        arrays,
        outletListenerSpecs,
        inletCallerSpecs,
        debug,
    };
};
/** Helper to get code macros from compile target. */
const getMacros = (target) => ({ javascript: macros, assemblyscript: macros$1 }[target]);
/** Helper to execute compilation */
const executeCompilation = (compilation) => {
    preCompileSignalAndMessageFlow(compilation);
    if (compilation.target === 'javascript') {
        return compileToJavascript(compilation);
    }
    else if (compilation.target === 'assemblyscript') {
        return compileToAssemblyscript(compilation);
    }
    else {
        throw new Error(`Invalid compilation.target ${compilation.target}`);
    }
};
class InvalidSettingsError extends Error {
}

export { index as default, executeCompilation, getMacros, validateSettings };
