import { mapArray } from '../functional-helpers.js';
import { getInlet, getNode, getOutlet } from './getters.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * Breadth first traversal for signal in the graph.
 * Traversal path is calculated by pulling incoming connections from
 * {@link nodesPullingSignal}.
 */
const signalNodes = (graph, nodesPullingSignal) => {
    const traversal = [];
    nodesPullingSignal.forEach((node) => _signalNodesBreadthFirstRecursive(traversal, [], graph, node));
    return traversal;
};
const _signalNodesBreadthFirstRecursive = (traversal, currentPath, graph, node) => {
    const nextPath = [...currentPath, node.id];
    Object.entries(node.sources)
        .filter(([inletId]) => getInlet(node, inletId).type === 'signal')
        .forEach(([_, sources]) => {
        sources.forEach((source) => {
            const sourceNode = getNode(graph, source.nodeId);
            if (currentPath.indexOf(sourceNode.id) !== -1) {
                return;
            }
            _signalNodesBreadthFirstRecursive(traversal, nextPath, graph, sourceNode);
        });
    });
    if (traversal.indexOf(node.id) === -1) {
        traversal.push(node.id);
    }
};
/**
 * Breadth first traversal for signal in the graph.
 * Traversal path is calculated by pulling incoming connections from
 * {@link nodesPushingMessages}.
 */
const messageNodes = (graph, nodesPushingMessages) => {
    const traversal = [];
    nodesPushingMessages.forEach((node) => {
        _messageNodesDepthFirstRecursive(traversal, graph, node);
    });
    return traversal;
};
const _messageNodesDepthFirstRecursive = (traversal, graph, node) => {
    if (traversal.indexOf(node.id) !== -1) {
        return;
    }
    traversal.push(node.id);
    Object.entries(node.sinks)
        .filter(([outletId]) => getOutlet(node, outletId).type === 'message')
        .forEach(([_, sinks]) => {
        sinks.forEach((sink) => {
            _messageNodesDepthFirstRecursive(traversal, graph, getNode(graph, sink.nodeId));
        });
    });
};
/**
 * Remove dead sinks and sources in graph.
 * @param graphTraversal contains all nodes that are connected to
 * an input or output of the graph.
 */
const trimGraph = (graph, graphTraversal) => mapArray(Object.values(graph).filter((node) => graphTraversal.includes(node.id)), (node) => [
    node.id,
    {
        ...node,
        sources: removeDeadSources(node.sources, graphTraversal),
        sinks: removeDeadSinks(node.sinks, graphTraversal),
    },
]);
/**
 * When `node` has a sink node that is not connected to an end sink, that sink node won't be included
 * in the traversal, but will still appear in `node.sinks`.
 * Therefore, we need to make sure to filter `node.sinks` to exclude sink nodes that don't
 * appear in the traversal.
 */
const removeDeadSinks = (sinks, graphTraversal) => {
    const filteredSinks = {};
    Object.entries(sinks).forEach(([outletId, outletSinks]) => {
        const filteredOutletSinks = outletSinks.filter(({ nodeId: sinkNodeId }) => graphTraversal.includes(sinkNodeId));
        if (filteredOutletSinks.length) {
            filteredSinks[outletId] = filteredOutletSinks;
        }
    });
    return filteredSinks;
};
/**
 * Filters a node's sources to exclude source nodes that don't
 * appear in the traversal.
 */
const removeDeadSources = (sources, graphTraversal) => {
    const filteredSources = {};
    Object.entries(sources).forEach(([inletId, inletSources]) => {
        const filteredInletSources = inletSources.filter(({ nodeId: sourceNodeId }) => graphTraversal.includes(sourceNodeId));
        if (filteredInletSources.length) {
            filteredSources[inletId] = filteredInletSources;
        }
    });
    return filteredSources;
};

export { messageNodes, removeDeadSinks, removeDeadSources, signalNodes, trimGraph };
