import { mapObject, mapArray } from '../../functional-helpers.js';
import { liftString, readTypedArray } from './core-bindings.js';
import { lowerMessage, liftMessage } from './msg-bindings.js';
import { instantiateWasmModule } from './wasm-helpers.js';

/*
 * Copyright (c) 2022-2023 Sébastien Piquemal <sebpiq@protonmail.com>, Chris McCormick.
 *
 * This file is part of WebPd
 * (see https://github.com/sebpiq/WebPd).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
// This must be called again when doing something on the wasm module
// which could cause memory grow (lowerString, lowerMessage,
//      lowerBuffer, lowerMessage) :
// https://github.com/emscripten-core/emscripten/issues/6747
const updateWasmInOuts = (rawModule, engineData) => {
    engineData.wasmOutput = readTypedArray(rawModule, engineData.arrayType, rawModule.getOutput());
    engineData.wasmInput = readTypedArray(rawModule, engineData.arrayType, rawModule.getInput());
};
const createEngineLifecycleBindings = (rawModule, engineData) => {
    return {
        configure: {
            type: 'proxy',
            value: (sampleRate, blockSize) => {
                engineData.metadata.audioSettings.blockSize = blockSize;
                engineData.metadata.audioSettings.sampleRate = sampleRate;
                engineData.blockSize = blockSize;
                rawModule.configure(sampleRate, blockSize);
                updateWasmInOuts(rawModule, engineData);
            },
        },
        loop: {
            type: 'proxy',
            value: (input, output) => {
                for (let channel = 0; channel < input.length; channel++) {
                    engineData.wasmInput.set(input[channel], channel * engineData.blockSize);
                }
                updateWasmInOuts(rawModule, engineData);
                rawModule.loop();
                updateWasmInOuts(rawModule, engineData);
                for (let channel = 0; channel < output.length; channel++) {
                    output[channel].set(engineData.wasmOutput.subarray(engineData.blockSize * channel, engineData.blockSize * (channel + 1)));
                }
            },
        },
    };
};
const createInletCallersBindings = (rawModule, engineData) => mapObject(engineData.metadata.compilation.inletCallerSpecs, (inletIds, nodeId) => ({
    type: 'proxy',
    value: mapArray(inletIds, (inletId) => [
        inletId,
        (message) => {
            const messagePointer = lowerMessage(rawModule, message);
            rawModule[engineData.metadata.compilation.codeVariableNames
                .inletCallers[nodeId][inletId]](messagePointer);
        },
    ]),
}));
const createOutletListenersBindings = (_, engineData) => mapObject(engineData.metadata.compilation.outletListenerSpecs, (outletIds) => ({
    type: 'proxy',
    value: mapArray(outletIds, (outletId) => [
        outletId,
        {
            onMessage: () => undefined,
        },
    ]),
}));
const outletListenersImports = (forwardReferences, metadata) => {
    const wasmImports = {};
    const { codeVariableNames } = metadata.compilation;
    Object.entries(metadata.compilation.outletListenerSpecs).forEach(([nodeId, outletIds]) => {
        outletIds.forEach((outletId) => {
            const listenerName = codeVariableNames.outletListeners[nodeId][outletId];
            wasmImports[listenerName] = (messagePointer) => {
                const message = liftMessage(forwardReferences.rawModule, messagePointer);
                forwardReferences.modules.outletListeners[nodeId][outletId].onMessage(message);
            };
        });
    });
    return wasmImports;
};
const readMetadata = async (wasmBuffer) => {
    // In order to read metadata, we need to introspect the module to get the imports
    const inputImports = {};
    const wasmModule = WebAssembly.Module.imports(new WebAssembly.Module(wasmBuffer));
    // Then we generate dummy functions to be able to instantiate the module
    wasmModule
        .filter((imprt) => imprt.module === 'input' && imprt.kind === 'function')
        .forEach((imprt) => (inputImports[imprt.name] = () => undefined));
    const wasmInstance = await instantiateWasmModule(wasmBuffer, {
        input: inputImports,
    });
    // Finally, once the module instantiated, we read the metadata
    const wasmExports = wasmInstance.exports;
    const stringPointer = wasmExports.metadata.valueOf();
    const metadataJSON = liftString(wasmExports, stringPointer);
    return JSON.parse(metadataJSON);
};

export { createEngineLifecycleBindings, createInletCallersBindings, createOutletListenersBindings, outletListenersImports, readMetadata, updateWasmInOuts };
